#include "config.h"

#include <boost/algorithm/string.hpp>
#include <cctype>
#include <fstream>
#include <iostream>
#include <string>
#include <string_view>

#include "json5cpp.h"
#include "log.h"

namespace tools {
bool ends_with(std::string_view str, std::string_view suffix) {
  return str.size() >= suffix.size() &&
         str.compare(str.size() - suffix.size(), suffix.size(), suffix) == 0;
}
}  // namespace tools

template <>
uint32_t ConfigMap::Get<uint32_t>(const std::string &k) {
  ValueType value = at(k);
  try {
    return value.asUInt();
  } catch (std::exception &e1) {
    const auto str_val = value.asString();
    LOG(DEBUG) << "Key: " << k << "| Value '" << str_val << "' | Exception: " << e1.what();
    LOG(DEBUG) << "Recovering by converting to uint32_t";
    try {
      return static_cast<uint32_t>(std::stoul(str_val));
    } catch (std::exception &e2) {
      LOG(ERROR) << "Conversion of '" << str_val << "' failed with exception: " << e2.what();
      throw;
    }
  }
}

template <>
uint64_t ConfigMap::Get<uint64_t>(const std::string &k) {
  ValueType value = at(k);
  try {
    return value.asUInt64();
  } catch (std::exception &e1) {
    const auto str_val = value.asString();
    LOG(DEBUG) << "Key: " << k << "| Value '" << str_val << "' | Exception: " << e1.what();
    LOG(DEBUG) << "Recovering by converting to uint64_t";
    try {
      return static_cast<uint64_t>(std::stoul(str_val));
    } catch (std::exception &e2) {
      LOG(ERROR) << "Conversion of '" << str_val << "' failed with exception: " << e2.what();
      throw;
    }
  }
}

template <>
int ConfigMap::Get<int>(const std::string &k) {
  ValueType value = at(k);
  try {
    return value.asInt();
  } catch (std::exception &e1) {
    const auto str_val = value.asString();
    LOG(DEBUG) << "Key: " << k << "| Value '" << str_val << "' | Exception: " << e1.what();
    LOG(DEBUG) << "Recovering by converting to int";
    try {
      return static_cast<int>(std::stoul(str_val));
    } catch (std::exception &e2) {
      LOG(ERROR) << "Conversion of '" << str_val << "' failed with exception: " << e2.what();
      throw;
    }
  }
}

template <>
bool ConfigMap::Get<bool>(const std::string &k) {
  std::string str = at(k).asString();
  boost::algorithm::to_lower(str);
  return str == "true" || str == "yes";
}

template <>
int ConfigMap::Parse<ConfigMap::EFormat::JSON5>(std::ifstream &input) {
  ValueType value;
  std::string err;
  if (Json5::parse(input, value, &err)) {
    Import(value);
  } else {
    LOG(ERROR) << "Invalid JSON5: " << err;
    return -1;
  }
  return 0;
}

// NOTE: Support for "conf" files following the format below is deprecated. Legacy "conf" parser to
// be dropped once JSON-based pipelines become the standard.
template <>
int ConfigMap::Parse<ConfigMap::EFormat::CONF>(std::ifstream &in) {
  bool valid = true;

  std::string line;
  while (!std::getline(in, line).eof()) {
    // Perform left trim
    line.erase(line.begin(),
               std::find_if(line.begin(), line.end(), [](int ch) { return !std::isspace(ch); }));

    // Skip empty lines
    if (line.empty()) {
      continue;
    }

    // Skip comments
    if (line.size() > 1 && line[0] == '#') {
      continue;
    }

    auto delim = line.find_first_of(':');
    if (delim == std::string::npos) {
      // Skip ill-formatted lines
      continue;
    }

    std::string key = line.substr(0, delim);
    std::string value = line.substr(delim + 1);

    if (key.empty()) {
      // Skip ill-formatted lines
      continue;
    }

    if (value.empty()) {
      LOG(ERROR) << "Configuration entry " << key << " has no value!";
      valid &= false;
    }

    Insert(key, value);
  }

  if (!valid) {
    throw std::invalid_argument("Configuration error: Keys with no value detected!");
  }

  return 0;
}

Config::Config(std::string_view filename) {
  const auto filename_str = std::string(filename);
  // Open file and check if valid
  std::ifstream in(filename_str, std::ios_base::in);
  if (!in.is_open()) {
    throw std::invalid_argument("Configuration file (" + filename_str +
                                ") could not be opened. Does it exist?");
  }

  // Pass open file handle to parser according to filename ending
  if (tools::ends_with(filename_str, ".json5")) {
    dict_.Parse<ConfigMap::EFormat::JSON5>(in);
  } else if (tools::ends_with(filename_str, ".conf")) {
    dict_.Parse<ConfigMap::EFormat::CONF>(in);
  } else {
    throw std::invalid_argument("Configuration file (" + filename_str +
                                ") has unrecognized suffix. Please specify the intended format.");
  }

  // Setup configs
  input_type_ = ToInputType(dict_.Get("input"));
  input_file_ = dict_.Get("input_file");
  scone_address_ =
      SconeAddress(dict_.Get("scone_host"), dict_.Get("scone_port"), dict_.Get("scone_board"));
  dma_device_ = dict_.Get("dma_dev");
  attempt_tcp_reconnect_ = dict_.Get<bool>("dev_TCPAutoReconnectOnFailure");

  port_ = dict_.Get<int>("port"),

  num_threads_ = dict_.Get<uint32_t>("threads");
  max_file_size_ = dict_.Get<uint64_t>("max_file_size");

  num_packets_per_report_ = dict_.Get<uint32_t>("packets_per_report");
  num_packet_buffers_ = dict_.Get<uint64_t>("dma_number_of_packet_buffers");
  packet_buffer_size_ = dict_.Get<uint64_t>("dma_packet_buffer_size");
  min_log_severity_ = ToLOG_LEVEL(dict_.Get("log_min_severity"));

  // Run configs
  tcp_port_ = dict_.Get<uint32_t>("tcpDestPort");
  num_orbits_per_packet_ = dict_.Get<uint64_t>("nOrbitsPerPacket");

  output_file_base_ = dict_.Get("output_filename_base");
  output_file_prefix_ = dict_.Get("output_filename_prefix");
  // num_orbits_per_file_: if > 0, use fixed N orbits per file
  num_orbits_per_file_ = dict_.Get<uint64_t>("nOrbitsPerFile");
  force_write_out_ = dict_.Get<bool>("output_force_write");

  processor_type_ = ToProcessorType(dict_.Get("processor_type"));
  source_id_ = GetSourceID(processor_type_);
  enable_processing_ = dict_.Get<bool>("enable_stream_processor");
  apply_zero_suppression_ = dict_.Get<bool>("doZS");
  support_cmssw_headers_ = dict_.Get<bool>("cmsswHeaders");
}

void Config::print() const { dict_.Print(); }