#ifndef CONFIG_H
#define CONFIG_H

#include <json/json.h>

#include <boost/algorithm/string.hpp>
#include <cstdint>
#include <map>
#include <stdexcept>
#include <string>
#include <string_view>

#include "log.h"

class ConfigMap {
 public:
  enum class EFormat { CONF, JSON5 };

  using ValueType = Json::Value;
  using ConstIteratorType = Json::Value::const_iterator;

  template <EFormat>
  int Parse(std::ifstream &input);

  // Insert (along with other functions) takes `key` as a const-ref since `ValueType`'s operator[]
  // does not support `std::string_view`.
  void Insert(const std::string &key, ValueType val) { data_[key] = std::move(val); }

  void Import(const ValueType &values) { data_ = values; }

  void Print() const {
    for (ConstIteratorType it = data_.begin(); it != data_.end(); it++) {
      auto k = it.key().asString();
      LOG(INFO) << "key: " << k << " | value: " << data_[k];
    }
  }

  template <typename T = std::string>
  T Get(const std::string &key) {
    return at(key).asString();
  }

 private:
  ValueType data_;

  inline ValueType at(const std::string &key) const {
    try {
      auto &val = data_[key];
      return val;
    } catch (std::exception &e) {
      LOG(ERROR) << "Error fetching value for key '" << key << "'";
      LOG(ERROR) << e.what();
      throw;
    }
  }
};

class Config {
 public:
  enum class InputType { WZDMA, FILEDMA, TCPIP, MICRONDMA, FILE };
  enum class ProcessorType { PASS_THROUGH, GMT, CALO, BRIL };

  struct SconeAddress {
    std::string host;
    std::string port;
    std::string board;

    SconeAddress() = default;
    SconeAddress(std::string_view h, std::string_view p, std::string_view b)
        : host(h), port(p), board(b) {
      assert(!host.empty() && !port.empty() && !board.empty());
    }

    friend std::ostream &operator<<(std::ostream &os, const SconeAddress &scone) {
      os << scone.host << ":" << scone.port << "/" << scone.board;
      return os;
    }
  };

  explicit Config(std::string_view filename);
  void print() const;

  ////// Setup configurations ////////////
  InputType input_type_;
  SconeAddress scone_address_;
  std::string tcp_dest_ports_range_;
  std::string dma_device_;
  bool attempt_tcp_reconnect_{false};
  uint32_t port_;
  uint32_t source_id_;
  uint32_t num_threads_;

  // Log, Statistics
  uint32_t num_packets_per_report_;
  uint64_t num_packet_buffers_;
  uint64_t packet_buffer_size_;
  LOG_LEVEL min_log_severity_{LOG_LEVEL::ERROR};

  ////// Run configurations ////////////
  // Input
  std::string input_file_;
  uint32_t tcp_port_;
  uint32_t num_orbits_per_packet_;
  // Processor
  ProcessorType processor_type_;
  bool enable_processing_;
  bool apply_zero_suppression_;
  // Output
  std::string output_file_base_;
  std::string output_file_prefix_;

  uint32_t run_id_;
  bool force_write_out_{false};

  // File Stream outputs
  uint64_t max_file_size_;

  // Orbit Stream outputs
  bool support_cmssw_headers_;
  uint64_t num_orbits_per_file_;

  inline LOG_LEVEL ToLOG_LEVEL(const std::string &in) const {
    std::string str = boost::algorithm::to_upper_copy(in);
    if (str == "TRACE") return TRACE;
    if (str == "DEBUG") return DEBUG;
    if (str == "INFO") return INFO;
    if (str == "WARNING") return WARNING;
    if (str == "ERROR") return ERROR;
    if (str == "FATAL") return FATAL;
    throw std::invalid_argument("Configuration error: Unknown log level '" + str + "'");
  }

  inline InputType ToInputType(const std::string &in) const {
    std::string str = boost::algorithm::to_upper_copy(in);
    if (str == "WZDMA") return InputType::WZDMA;
    if (str == "FILEDMA") return InputType::FILEDMA;
    if (str == "TCPIP") return InputType::TCPIP;
    if (str == "MICRONDMA") return InputType::MICRONDMA;
    if (str == "FILE") return InputType::FILE;
    throw std::invalid_argument("Configuration error: Wrong input type '" + str + "'");
  }

  inline ProcessorType ToProcessorType(const std::string &in) const {
    std::string str = boost::algorithm::to_upper_copy(in);
    if (str == "PASS_THROUGH") return ProcessorType::PASS_THROUGH;
    if (str == "GMT") return ProcessorType::GMT;
    if (str == "CALO") return ProcessorType::CALO;
    if (str == "BRIL") return ProcessorType::BRIL;
    throw std::invalid_argument("Configuration error: Wrong processor type '" + str + "'");
  }

  inline uint32_t GetSourceID(ProcessorType processor) const {
    // Source ID is only used to fill the output CMSSW FRD event header, which is currently
    // supported for the GMT and Calo processors.
    switch (processor) {
      case ProcessorType::GMT:
        return 1;
      case ProcessorType::CALO:
        return 2;
      default:
        if (!support_cmssw_headers_) {
          LOG(WARNING)
              << "Specified `processor_type` has no pre-defined source ID. Using `source_id`=0.";
          return 0;
        } else {
          const char msg_incompatible[] =
              "Specified `processor_type` incompatible with CMSSW header-enabled outputs. Disable "
              "`cmsswHeaders` or change `processor_type` to one of {GMT, CALO}.";
          LOG(ERROR) << msg_incompatible;
          throw std::invalid_argument(msg_incompatible);
        }
    }
  }

 private:
  ConfigMap dict_;
};
#endif
