#include "pipeline.h"

#include "bril_processor.h"
#include "calo_orbit_processor.h"
#include "muon_orbit_processor.h"
#include "orbit_processor.h"
#include "processor.h"

template <typename T>
std::shared_ptr<OrbitProcessor> MakeA(const Config &conf, ctrl &control) {
  return std::make_shared<T>(conf.num_orbits_per_packet_, conf.enable_processing_,
                             conf.support_cmssw_headers_, static_cast<uint16_t>(conf.source_id_),
                             control);
}

std::shared_ptr<Processor> MakeProcessor(const Config &conf, ctrl &control) {
  switch (conf.processor_type_) {
    case Config::ProcessorType::GMT:
      return MakeA<MuonOrbitProcessor>(conf, control);
    case Config::ProcessorType::CALO:
      return MakeA<CaloOrbitProcessor>(conf, control);
    case Config::ProcessorType::BRIL:
      return std::make_shared<BrilProcessor>(control);
    case Config::ProcessorType::PASS_THROUGH:
      return std::make_shared<Processor>();
    default:
      throw std::invalid_argument("Error: Unrecognized processor type.");
  }
}

Pipeline::Pipeline(Config &conf, uint maxTokensPerThread)
    : fMaxTokensPerThread(maxTokensPerThread), conf_(conf) {
  tools::log::set_filter_min_severity(conf_.min_log_severity_);
  conf_.print();
  const Config::InputType &input = conf_.input_type_;
  const size_t &nbPacketBuffers = conf_.num_packet_buffers_;
  const size_t &packetBufferSize = conf_.packet_buffer_size_;

  if (input == Config::InputType::FILEDMA) {
    // Create FILE DMA reader
    fInputFilter = std::make_shared<FileDmaInputFilter>(conf_.input_file_, packetBufferSize,
                                                        nbPacketBuffers, control);

  } else if (input == Config::InputType::WZDMA) {
    // Create WZ DMA reader
    fInputFilter =
        std::make_shared<WZDmaInputFilter>(packetBufferSize, nbPacketBuffers, control, conf_);

  } else if (input == Config::InputType::MICRONDMA) {
    // create MicronDmaInputFilter reader
    fInputFilter =
        std::make_shared<MicronDmaInputFilter>(packetBufferSize, nbPacketBuffers, control, conf_);

  } else if (input == Config::InputType::TCPIP) {
    // create TcpInputFilter reader
    if (conf_.attempt_tcp_reconnect_) {
      fInputFilter =
          std::make_shared<TcpInputFilter>(conf_.tcp_port_, packetBufferSize, nbPacketBuffers,
                                           conf_.num_orbits_per_packet_, control);

    } else {
      throw std::invalid_argument(
          "Configuration error: enable developmentMode to use TCP input filter");
    }
  } else {
    throw std::invalid_argument("Configuration error: Unknown input type was specified");
  }

  // Add input reader to a pipeline
  fPipeline.add_filter(*fInputFilter);

  // Create reformatter and add it to the pipeline
  if (conf_.enable_processing_) {
    fStreamProcessor = MakeProcessor(conf_, control);
    fPipeline.add_filter(*fStreamProcessor);
  }

  // Create file-writing stage and add it to the pipeline
  if (conf_.num_orbits_per_file_) {
    fOutputStream = std::make_shared<OutputByOrbitStream>(control, conf_);
  } else {
    fOutputStream =
        std::make_shared<OutputBySizeStream>(conf_.output_file_base_, conf_.output_file_prefix_,
                                             Pipeline::control, conf_.processor_type_);
  }
  fPipeline.add_filter(*fOutputStream);

  //    tbb::tick_count mainStartTime = tbb::tick_count::now();
  control.running = false;
  control.run_number = 0;
  control.max_file_size = conf_.max_file_size_;  // in bytes
  control.packets_per_report = static_cast<int>(conf_.num_packets_per_report_);
  control.output_force_write = conf_.force_write_out_;
  control.n_orbits_per_packet = static_cast<int>(conf_.num_orbits_per_packet_);

  // Firmware needs at least 1MB buffer for DMA
  if (packetBufferSize < 1024 * 1024) {
    LOG(ERROR) << "dma_packet_buffer_size must be at least 1048576 bytes (1MB), but "
               << packetBufferSize << " bytes was given. Check the configuration file.";
    throw std::invalid_argument("Configuration error: DMA packet buffer size too small. ");
  }
}

MultiPipeline::MultiPipeline(std::vector<Config> &configs) {
  for (auto &config : configs) {
    try {
      fPipelines.emplace_back(config);
      LOG(DEBUG) << "-Configuration loaded";
      LOG(DEBUG) << "-Pipeline set up";
    } catch (std::invalid_argument &e) {
      LOG(FATAL) << "-Configuration invalid! Error text is \"" << e.what() << "\" Bailing out.";
      throw;
    } catch (std::exception &e) {
      LOG(ERROR) << "-Error occurred. error text is: \"" << e.what() << "\"";
      throw;
    }
  }

  {  // RAII
    fServer = std::unique_ptr<server>(
        new server(io_service, fPipelines[0].conf_.port_, Pipeline::control));
    fServerThread = boost::thread(boost::bind(&boost::asio::io_service::run, &io_service));
  }
}

int MultiPipeline::Run() {
  int retval = 0;
  assert(!fPipelines.empty());
  try {
    if (!RunPipelines(static_cast<unsigned int>(fPipelines.size()))) {
      retval = 1;  // Will terminate with error code set
    }
  } catch (std::exception &e) {
    LOG(ERROR) << "Error in pipelines. Error text is: \"" << e.what() << "\"";
    retval = 1;                             // Will terminate with error code set
    Config &conf = fPipelines.at(0).conf_;  // Use info from first file for SCONE reset
    LOG(INFO) << "Resetting board at http://" << conf.scone_address_ << " before exiting.. ";
    int res = tools::reset_board(conf.scone_address_.host, conf.scone_address_.port,
                                 conf.scone_address_.board);
    if (res != 0) {
      LOG(ERROR) << "Reset failed! Error code: " << res;
    } else {
      LOG(INFO) << "Successfully reset the board.";
    }
  }

  LOG(DEBUG) << "Terminating the internal TCP server.";
  io_service.stop();
  if (!fServerThread.try_join_for(boost::chrono::milliseconds(2000))) {
    LOG(ERROR) << "Failed to terminate the internal TCP server, the program "
                  "may crash on exit.";
  } else {
    LOG(DEBUG) << "Internal TCP server is terminated.";
  }

  return retval;
}

int MultiPipeline::RunPipelines(uint numStreams) {
  tbb::task_group group;
  // Run pipelines and measure real time
  tbb::tick_count t0 = tbb::tick_count::now();
  for (uint i = 0; i < numStreams; ++i) {
    auto &p = this->fPipelines.at(i);
    group.run([&] { return p.fPipeline.run(p.conf_.num_threads_ * p.fMaxTokensPerThread); });
  }
  group.wait();
  tbb::tick_count t1 = tbb::tick_count::now();

  if (fVerbose) {
    LOG(INFO) << "time = " << (t1 - t0).seconds();
  }
  return 1;
}